﻿
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Reflection;

namespace ReflectionPlay
{
    class Demo
    {
        static void Main(string[] args)
        {
            //The GetType method returns an instance of class Type,
            //which describes the runtime type (i.e. class) of an object.  
            //All classes inherit GetType from Object.
            string s = "foo";
            Type t = s.GetType();

            //You can access information about all aspects of a class through
            //reflection....

            //...its identity and base class,
            Console.WriteLine(t);
            Console.WriteLine(t.BaseType);
            
            //...its methods
            var ms = t.GetMethods();

            foreach (MethodInfo mi in ms)
            {
                Console.WriteLine(mi.ToString() + mi.Attributes);
            }

            //...and (not shown) fields, generic parameters, etc.

            Console.WriteLine("===========");


            //The following code illustrates looking-up and invoking
            //a method.
            foreach (var mi in ms)
            {
                if (mi.Name == "ToUpper" && mi.GetParameters().Length == 0)
                {
                    Console.WriteLine("Invoking ToUpper on s: " +
                                      mi.Invoke(s, new Object[0] { }));
                }
            }

            Console.WriteLine("============");

            //The following code passes parameters to GetMethods() in order to
            //access a class's private components.  The parameters are
            //enumerations.  For more information on enumerations, see
            // http://msdn.microsoft.com/en-us/library/sbbt4032.aspx
            Type tt = typeof(Silly);

            ms = tt.GetMethods(BindingFlags.Public | BindingFlags.NonPublic
                                                   | BindingFlags.Instance);

            Console.WriteLine(tt.Name);

            foreach (MethodInfo mi in ms)
            {
                Console.WriteLine("  " + mi.Name);

                //Beyond printing information about each method, we also
                //print its attributes.  Attributes are metadata that can
                //be associated with classes, methods, etc. and accessed
                //at runtime.  A method's builtin attributes include
                //privacy settings (public, private, family...), whether
                //the method is virtual, and more.
                //
                //Custom attributes, such as [Obsolete] or [Test]
                //can be defined and used to annotate methods (and classes) 
                //as illustrated in the definitions of Silly and 
                //TestAttribute, below.
                Console.WriteLine("    " + mi.Attributes);
                foreach(object o in mi.GetCustomAttributes(true))
                {
                    Console.WriteLine("    " + o.ToString());
                }
            }

            Console.WriteLine("============");


            //An assembly corresponds (roughly) to a single compilation
            //output: either a .dll library or a .exe executable.  The
            //following code loads and reflects over an executable, in this
            //case the program itself.
            Assembly an_assembly = Assembly.LoadFile(
                Environment.CurrentDirectory+"/ReflectionPlay.exe");
            Console.WriteLine(an_assembly + "\n");

            foreach (Type ttt in an_assembly.GetTypes())
            {
                Console.WriteLine(ttt.Name);
            }

            Console.WriteLine("======");

            //The Strange class provides a new, overloaded definition for 
            //GetType.  While this can confuse code trying to learn about 
            //Strange, you can call the expected implementation of GetType
            //by first casting through Object.  This works because
            //GetType is not virtual.
            var ss = new Strange();
            Console.WriteLine("{0} {1}", ss.GetType(),
                              ((object) ss).GetType());

            Console.ReadKey();
        }
    }

    public class Silly
    {
        public int Z { get; private set; }

        private void Foo() {}

        [Test]
        [Obsolete]
        public int add(int x, int y) { return x+y+Z; }

        public Silly() { Z = 399-005; }
    }

    public class TestAttribute : Attribute
    { }

    public class Strange
    {
        public new Type GetType() { return typeof(Int32); }
    }

}
